"""Q8 fetcher for pybraendstofpriser."""

from __future__ import annotations
import logging

from ..exceptions import ErrorFetchingProduct
from ..const import DIESEL, OCTANE_95
from ..tools import clean_product_name, clean_value, get_html_soup, get_website

baseurl = "https://www.q8.dk/paa-stationen/braendstof/"

PRODUCTS = {
    DIESEL: {"name": "Q8 GoEasy Diesel"},
    OCTANE_95: {"name": "Q8 GoEasy 95 E10"},
}

COMPANY_NAME = "Q8"

_LOGGER = logging.getLogger(__name__)


class FuelCompany:
    """Fuel company class."""

    async def fetch_price(self, product: str) -> float | None:
        """Fetch fuel prices."""
        return await self._parser(product)

    async def list_products(self) -> list[str]:
        """List available fuel products."""
        retlist = []
        for _, productDict in PRODUCTS.items():
            retlist.append(productDict["name"])
        return retlist

    async def _parser(self, product) -> float | None:
        """Parse the fetched data."""
        r = await get_website(baseurl, timeout=5)
        html = get_html_soup(r)
        rows = html.find_all("li", class_="price-item")

        for row in rows:
            cells = row.text.split("\r")
            if cells:
                found = PRODUCTS[product]["name"] == clean_product_name(cells[0])
                if found:
                    return clean_value(cells[1])

        raise ErrorFetchingProduct(f"Product '{PRODUCTS[product]['name']}' not found")



"""Defines the pybraendstofpriser package."""

from __future__ import annotations
from asyncio import get_running_loop
from collections import namedtuple
from genericpath import isfile
import importlib
import logging
from os import listdir
from posixpath import dirname
import sys

if sys.version_info < (3, 11, 0):
    sys.exit("The pybraendstofpriser module requires Python 3.11.0 or later")

_LOGGER = logging.getLogger(__name__)
Company = namedtuple("Company", "module namespace products name")


class Braendstofpriser:
    """Main class for pybraendstofpriser module."""

    def __init__(self):
        """Initialize the Braendstofpriser class."""
        # self.companies = []
        self.companies = {}
        _LOGGER.debug("Braendstofpriser initialized")

    async def list_companies(self):
        """List fuel companies."""
        _LOGGER.debug("Listing companies")
        loop = get_running_loop()
        companies = await loop.run_in_executor(
            None, listdir, f"{dirname(__file__)}/companies"
        )
        for company in sorted(companies):
            company_path = f"{dirname(__file__)}/companies/{company}"
            if (
                isfile(company_path)
                and not company.endswith("__pycache__")
                and not company == "__init__.py"
            ):
                company_name = company.replace(".py", "")
                _LOGGER.debug("Found company: %s", company_name)

                ns = f".companies.{company.replace('.py', '')}"
                mod = await self._load_module(ns)

                self.companies.update(
                    {
                        mod.COMPANY_NAME: {
                            "products": mod.PRODUCTS,
                            "namespace": ns,
                        }
                    }
                )

        return self.companies

    async def get_price(self, company: str, product: str):
        """Get fuel price for a specific company and product."""
        _LOGGER.debug("Getting price for %s - %s", company, product)
        c = await self._load_module(self.companies[company]["namespace"])
        co = c.FuelCompany()
        return await co.fetch_price(product)

    @staticmethod
    async def _load_module(namespace: str):
        """Dynamically load a module."""
        loop = get_running_loop()
        return await loop.run_in_executor(
            None, importlib.import_module, namespace, __name__
        )



"""Tools package for pybraendstofpriser."""

from __future__ import annotations
import logging
import aiohttp
from bs4 import BeautifulSoup as BS

_LOGGER = logging.getLogger(__name__)


@staticmethod
async def get_website(
    url: str,
    timeout: int = 10,
    headers: dict | None = None,
    as_json: bool = False,
    ssl_context=None,
):
    """Fetch content from a website asynchronously."""
    try:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(url, timeout=timeout, ssl=ssl_context) as response:
                response.raise_for_status()
                return await response.text() if not as_json else await response.json()
    except aiohttp.ClientError as e:
        _LOGGER.error("Error fetching %s: %s", url, e)
        return None


@staticmethod
def get_html_soup(r, parser="html.parser"):
    """Parse HTML content using BeautifulSoup."""
    return BS(r, parser)


@staticmethod
def clean_product_name(productName):
    """Clean and standardize product name."""
    productName = productName.replace("Beskrivelse: ", "")
    productName = productName.strip()
    return productName


@staticmethod
def clean_value(value) -> float | None:
    """Clean and convert value to float."""
    if isinstance(value, (float)):
        return value

    value = value.replace("kr.", "").replace(",", ".").strip()
    try:
        return float(value)
    except ValueError:
        _LOGGER.error("Error converting value to float: %s", value)
        return None


@staticmethod
async def download_file(
    url: str, filename: str, path: str, headers: dict | None = None, ssl_context=None
):
    """Download a file asynchronously using aiohttp."""
    try:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(url, ssl=ssl_context) as response:
                response.raise_for_status()
                full_path = path + filename
                with open(full_path, "wb") as file:
                    async for chunk in response.content.iter_chunked(1024):
                        file.write(chunk)
    except aiohttp.ClientError as e:
        _LOGGER.error("Error downloading file from %s: %s", url, e)



"""Go'On fetcher for pybraendstofpriser."""

from __future__ import annotations
import logging
import re
import subprocess

from ..exceptions import ErrorFetchingProduct
from ..const import DIESEL, DOWNLOAD_PATH, OCTANE_92, OCTANE_95
from ..tools import (
    clean_value,
    download_file,
    get_html_soup,
    get_website,
)

host = "https://goon.nu"
baseurl = f"{host}/wp-content/themes/goon/build/images"

PRODUCTS = {
    OCTANE_92: {"name": "Blyfri 92", "ocr_crop": ["58", "176", "134", "46"]},
    OCTANE_95: {"name": "Blyfri 95", "ocr_crop": ["58", "232", "134", "46"]},
    DIESEL: {"name": "Diesel", "ocr_crop": ["58", "289", "134", "46"]},
}

COMPANY_NAME = "Go'On"

_LOGGER = logging.getLogger(__name__)


class FuelCompany:
    """Fuel company class."""

    async def fetch_price(self, product: str) -> float | None:
        """Fetch fuel prices."""
        return await self._parser(product)

    async def list_products(self) -> list[str]:
        """List available fuel products."""
        retlist = []
        for _, productDict in PRODUCTS.items():
            retlist.append(productDict["name"])
        return retlist

    async def _parser(self, product) -> float | None:
        """Parse the fetched data."""
        r = await get_website(baseurl, timeout=5)
        html = get_html_soup(r)
        file = [
            host + node.get("href")
            for node in html.find_all(
                "a", {"href": re.compile(r"[\w]{5,}[\d]{10,}.png")}
            )
        ][0]

        await download_file(file, "goon.png", DOWNLOAD_PATH)

        ocr_cmd = (
            ["ssocr"]
            + ["-d5"]
            + ["-t20"]
            + ["make_mono", "invert", "-D"]
            + ["crop"]
            + PRODUCTS[product]["ocr_crop"]
            + [DOWNLOAD_PATH + "goon.png"]
        )
        # Perform OCR on the cropped image
        with subprocess.Popen(
            ocr_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as ocr:
            out = ocr.communicate()
            if out[0] != b"":
                return clean_value(out[0].strip().decode("utf-8"))

        # raise ErrorFetchingProduct(f"Product '{PRODUCTS[product]['name']}' not found")



"""Exceptions for pybraendstofpriser."""


class ErrorFetchingData(Exception):
    """Exception raised for errors in fetching data from the website."""


class ErrorFetchingProduct(Exception):
    """Exception raised for errors in fetching a specific product."""



"""Shell fetcher for pybraendstofpriser."""

from __future__ import annotations
import logging
import ssl

from ..exceptions import ErrorFetchingProduct
from ..const import DIESEL, OCTANE_100, OCTANE_95, DIESEL_PLUS
from ..tools import clean_product_name, clean_value, get_html_soup, get_website

baseurl = "https://shellservice.dk/wp-json/shell-wp/v2/daily-prices"

PRODUCTS = {
    DIESEL: {"name": "Shell FuelSave Diesel", "ProductCode": "315000"},
    DIESEL_PLUS: {"name": "Shell V-Power Diesel", "ProductCode": "325000"},
    OCTANE_95: {"name": "Shell FuelSave 95 oktan", "ProductCode": "475000"},
    OCTANE_100: {"name": "Shell V-Power 100 oktan", "ProductCode": "465000"},
}

COMPANY_NAME = "Shell"

_LOGGER = logging.getLogger(__name__)

# Create the SSL context globally to reuse it
SSL_CONTEXT = ssl.create_default_context()
SSL_CONTEXT.minimum_version = ssl.TLSVersion.TLSv1_3
SSL_CONTEXT.maximum_version = ssl.TLSVersion.TLSv1_3


class FuelCompany:
    """Fuel company class."""

    async def fetch_price(self, product: str) -> float | None:
        """Fetch fuel prices."""
        return await self._parser(product)

    async def list_products(self) -> list[str]:
        """List available fuel products."""
        retlist = []
        for _, productDict in PRODUCTS.items():
            retlist.append(productDict["name"])
        return retlist

    async def _parser(self, product) -> float | None:
        """Parse the fetched data."""
        headers = {"User-Agent": "pybraendstofpriser"}
        r = await get_website(
            baseurl, timeout=5, as_json=True, headers=headers, ssl_context=SSL_CONTEXT
        )
        prod_data = r["results"]["products"]

        for item in prod_data:
            if item["id"] == PRODUCTS[product]["ProductCode"]:
                return clean_value(item["price_incl_vat"])

        raise ErrorFetchingProduct(f"Product '{PRODUCTS[product]['name']}' not found")
